/*
 * GameText.cs - Manages a text buffer and provides methods necessary for text
 *               output and shell functionality in a Level 9 game.
 *
 * Copyright (C) 2004 - 2011 Andreas Scherrer
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this software; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 * USA.
 */

using System;
using System.Drawing;
using System.Collections.Generic;
using System.Windows.Forms;

namespace Level9
{
	/// <summary>
	/// Manages a text buffer and provides methods necessary for text output
	/// and shell functionality in a Level 9 game.
	/// </summary>
	public class GameText
	{
		// -- Constants ------------------------------------------------------

		/// <summary>
		/// The maximum number of commands for the command history.
		/// </summary>
		const int MAX_COMMANDS_HISTORY = 100;

		/// <summary>
		/// The maximum length of user input.
		/// </summary>
		const int MAX_LENGTH_INPUT = 160;

		// -- Attributes -----------------------------------------------------

		/// <summary>
		/// The list of paragraphs associated with this GameText instance.
		/// </summary>
		List<AttributedText> paragraphs = new List<AttributedText>();

		/// <summary>
		/// A list of output lines. Stores the text provided by the last call to
		/// GetOutput().
		/// </summary>
		List<AttributedText> output = new List<AttributedText>();

		/// <summary>
		/// The text offset position in the list of paragraphs.
		/// </summary>
		TextPosition offsetPos = new TextPosition();

		/// <summary>
		/// The caret position in the list of paragraphs.
		/// </summary>
		TextPosition caretPos = new TextPosition();

		/// <summary>
		/// User input position in the list of paragraphs.
		/// </summary>
		TextPosition inputPos = new TextPosition();

		/// <summary>
		/// The command history.
		/// </summary>
		List<string> commandHistory = new List<string>();

		/// <summary>
		/// The current position in the command history.
		/// </summary>
		int commandHistoryPos = 0;

		/// <summary>
		/// The text viewer component that displays the scrollback buffer.
		/// </summary>
		TextForm scrollbackView = null;

		// -- Constructors ---------------------------------------------------

		/// <summary>
		/// Initializes a new instance of the GamePanel class.
		/// </summary>
		public GameText()
		{
			scrollbackView = new TextForm(false,Config.Instance.TextViewerSize);
			scrollbackView.Text = "Scrollback";
		}

		// -- Methods --------------------------------------------------------

		/// <summary>
		/// Appends text to the list of paragraphs.
		/// </summary>
		/// <param name="text">the text to append</param>
		public void Append(string text)
		{
			if (text == null || text.Length == 0) {
				return;
			}

			string[] split = text.Split(new char[] {'\r'});
			for (int i = 0; i < split.Length; i++) {
				if (paragraphs.Count > 0 && i == 0) {
					paragraphs[paragraphs.Count - 1].Append(split[i]);
				}
				else {
					paragraphs.Add(new AttributedText(split[i]));
				}
			}

			if (!scrollbackView.IsDisposed) {
				scrollbackView.AppendText(text, SystemColors.WindowText);
				scrollbackView.SetCaretPosition(
					scrollbackView.GetTextLength() - 1);
			}
		}

		/// <summary>
		/// Provides output lines for the given number of lines and columns.
		/// </summary>
		/// <param name="lines">the number of lines</param>
		/// <param name="columns">the number of columns per line</param>
		/// <param name="moreLines">
		/// true if more lines are avialable, else false
		/// </param>
		/// <param name="changeOffset">
		/// provides the offset at which lines have changed in comparison to the
		/// last GetOutput() call.
		/// </param>
		/// <returns>a list of output lines</returns>
		public List<AttributedText> GetOutput(int lines, int columns,
			out bool moreLines, out int changeOffset)
		{
			//
			// Create output lines
			//
			List<AttributedText> newOutput = TextUtils.SplitParagraphs(
				paragraphs, lines + 1, columns, offsetPos, true);

			if (newOutput.Count > lines) {
				moreLines = true;
				newOutput = newOutput.GetRange(0, lines);
			}
			else {
				moreLines = false;
				newOutput = TextUtils.SplitParagraphs(paragraphs, lines,
					columns, false);
			}
			//
			// Determine the offset at which lines have changed
			//
			changeOffset = 0;
			for (int i = 0; i < newOutput.Count && i < output.Count; i++) {
				if (newOutput[i].Equals(output[i])) {
					changeOffset = i;
				}
				else {
					break;
				}
			}
			output = newOutput;

			return newOutput;
		}

		/// <summary>
		/// Sets the text offset position to the last character in the list of
		/// output lines.
		/// </summary>
		public void SetTextOffset()
		{
			if (output != null && output.Count > 0) {
				offsetPos = output[output.Count - 1].Origin;
			}
		}

		/// <summary>
		/// Sets the caret and input position to the last character in the list
		/// of paragraphs.
		/// </summary>
		public void SetInputPosition()
		{
			int last = paragraphs.Count - 1;
			if (paragraphs.Count == 0 || inputPos.Paragraph == last) {
				return;
			}

			inputPos.Paragraph = last;
			inputPos.Column = paragraphs[last].GetLength();
			paragraphs[last].Append(" ");
			paragraphs[last].AddAttribute(AttributedText.ATTR_INPUT_START,
				inputPos.Column);

			ClearCaretInParagraphs();
			caretPos.Copy(inputPos);
			SetCaretInParagraphs();
		}

		/// <summary>
		/// Provides user input corresponding to the input text position.
		/// </summary>
		/// <returns>
		/// an input line, or null if the list of paragraphs is empty
		/// </returns>
		public string GetInput()
		{
			if (paragraphs.Count == 0) {
				return null;
			}

			ClearCaretInParagraphs();

			AttributedText paragraph = paragraphs[inputPos.Paragraph];
			paragraph.TrimEnd();
			int last = paragraph.GetLength() - 1;
			paragraph.AddAttribute(AttributedText.ATTR_INPUT_END, last);
			AttributedText input = paragraph.GetRange(inputPos.Column, last);

			if (input != null) {
				if (commandHistory.Count == MAX_COMMANDS_HISTORY) {
					commandHistory.RemoveAt(0);
				}
				commandHistory.Add(input.ToString());
				commandHistoryPos = commandHistory.Count;

				scrollbackView.AppendText(input.ToString(),
					SystemColors.HotTrack);
				scrollbackView.SetCaretPosition(
					scrollbackView.GetTextLength() - 1);
			}

			return (input == null ? "" : input.ToString());
		}

		/// <summary>
		/// Inserts a character at the caret position.
		/// </summary>
		/// <param name="character">the character to insert</param>
		public void CaretIns(char character)
		{
			CaretIns(character.ToString());
		}

		/// <summary>
		/// Inserts text at the caret position.
		/// </summary>
		/// <param name="text">the text to insert</param>
		public void CaretIns(string text)
		{
			if (paragraphs.Count == 0 || text == null || text.Length == 0) {
				return;
			}

			AttributedText paragraph = paragraphs[inputPos.Paragraph];
			int inputLength = paragraph.GetLength() - inputPos.Column - 1;
			int insertLength = 0;
			if (inputLength + text.Length <= MAX_LENGTH_INPUT) {
				insertLength = text.Length;
				paragraph.Insert(text, caretPos.Column);
			}
			else if (inputLength < MAX_LENGTH_INPUT) {
				insertLength = MAX_LENGTH_INPUT - inputLength;
				paragraph.Insert(text.Substring(0, insertLength),
					caretPos.Column);
			}

			ClearCaretInParagraphs();
			caretPos.Column += insertLength;
			SetCaretInParagraphs();
		}

		/// <summary>
		/// Sets the caret to the home position.
		/// </summary>
		public void CaretHome()
		{
			if (paragraphs.Count == 0) {
				return;
			}
			ClearCaretInParagraphs();
			caretPos.Column = inputPos.Column;
			SetCaretInParagraphs();
		}

		/// <summary>
		/// Sets the caret to the end position.
		/// </summary>
		public void CaretEnd()
		{
			if (paragraphs.Count == 0) {
				return;
			}
			ClearCaretInParagraphs();
			AttributedText paragraph = paragraphs[caretPos.Paragraph];
			caretPos.Column = paragraph.GetLength() - 1;
			SetCaretInParagraphs();
		}

		/// <summary>
		/// Moves the caret right.
		/// </summary>
		public void CaretRight()
		{
			if (paragraphs.Count == 0) {
				return;
			}
			ClearCaretInParagraphs();
			AttributedText paragraph = paragraphs[caretPos.Paragraph];
			if (caretPos.Column < paragraph.GetLength() - 1) {
				caretPos.Column += 1;
			}
			SetCaretInParagraphs();
		}

		/// <summary>
		/// Moves the caret left.
		/// </summary>
		public void CaretLeft()
		{
			if (paragraphs.Count == 0) {
				return;
			}
			ClearCaretInParagraphs();
			if (caretPos.Column > inputPos.Column) {
				caretPos.Column -= 1;
			}
			SetCaretInParagraphs();
		}

		/// <summary>
		/// Deletes a character at the caret position.
		/// </summary>
		public void CaretDelete()
		{
			if (paragraphs.Count == 0) {
				return;
			}
			AttributedText paragraph = paragraphs[caretPos.Paragraph];
			if (caretPos.Column < paragraph.GetLength() - 1) {
				paragraph.RemoveRange(caretPos.Column, caretPos.Column);
				SetCaretInParagraphs();
			}
		}

		/// <summary>
		/// Moves the caret left and deletes a character.
		/// </summary>
		public void CaretBack()
		{
			if (paragraphs.Count == 0) {
				return;
			}
			if (caretPos.Column > inputPos.Column) {
				AttributedText paragraph = paragraphs[caretPos.Paragraph];
				ClearCaretInParagraphs();
				caretPos.Column -= 1;
				paragraph.RemoveRange(caretPos.Column, caretPos.Column);
				SetCaretInParagraphs();
			}
		}

		/// <summary>
		/// Hides the caret.
		/// </summary>
		public void HideCaret()
		{
			ClearCaretInParagraphs();
		}

		/// <summary>
		/// Shows the caret.
		/// </summary>
		public void ShowCaret()
		{
			SetCaretInParagraphs();
		}

		/// <summary>
		/// Moves backward in the command history.
		/// </summary>
		public void HistoryUp()
		{
			if (commandHistoryPos > 0) {
				commandHistoryPos--;
				RecallCommand();
			}
		}

		/// <summary>
		/// Moves forward in the command history.
		/// </summary>
		public void HistoryDown()
		{
			if (commandHistoryPos < commandHistory.Count - 1) {
				commandHistoryPos++;
				RecallCommand();
			}
		}

		/// <summary>
		/// Recalls a command in the command history.
		/// </summary>
		void RecallCommand()
		{
			if (paragraphs.Count == 0) {
				return;
			}
			AttributedText paragraph = paragraphs[inputPos.Paragraph];
			paragraph.RemoveRange(inputPos.Column, paragraph.GetLength() - 1);
			inputPos.Clear();
			SetInputPosition();
			CaretIns(commandHistory[commandHistoryPos]);
		}

		/// <summary>
		/// Removes the caret at the position indicated by caretPos.
		/// </summary>
		void ClearCaretInParagraphs()
		{
			if (caretPos.Paragraph >= paragraphs.Count) {
				return;
			}
			paragraphs[caretPos.Paragraph].RemoveAttribute(
				AttributedText.ATTR_INPUT_CARET, caretPos.Column);
		}

		/// <summary>
		/// Sets the caret to the position indicated by caretPos.
		/// </summary>
		void SetCaretInParagraphs()
		{
			if (caretPos.Paragraph >= paragraphs.Count) {
				return;
			}
			paragraphs[caretPos.Paragraph].AddAttribute(
				AttributedText.ATTR_INPUT_CARET, caretPos.Column);
		}

		/// <summary>
		/// Displays a text viewer that represents the scollback buffer.
		/// </summary>
		/// <param name="parent">
		/// the dialog owner window to be set for the text viewer
		/// </param>
		public void ShowScrollback(IWin32Window parent)
		{
			scrollbackView.BringToFront();
			scrollbackView.Show();
		}

		/// <summary>
		/// Clears the text information.
		/// </summary>
		public void Clear()
		{
			paragraphs.Clear();
			output.Clear();
			offsetPos.Clear();
			inputPos.Clear();
			caretPos.Clear();
			commandHistory.Clear();
			scrollbackView.Clear();
			commandHistoryPos = 0;
		}

		/// <summary>
		/// Provides the number of paragraphs.
		/// </summary>
		/// <returns>the number of paragraphs</returns>
		public int GetParagraphCount()
		{
			return paragraphs.Count;
		}

		/// <summary>
		/// Provides the number of output lines.
		/// </summary>
		/// <returns>the number of output lines</returns>
		public int GetOutputLinesCount()
		{
			return output.Count;
		}

		// -- Accessors ------------------------------------------------------

		/// <summary>
		/// Gets the list of paragraphs associated with this GameText instance.
		/// </summary>
		public List<AttributedText> Paragraphs
		{
			get {
				return paragraphs;
			}
		}
	}
}
